# -*- coding:utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from copy import copy
from abc import ABCMeta, abstractmethod
from collections import OrderedDict, Iterable, Mapping

import six

from .fields import BaseSchemaElement
from .fields.utils import is_iterable

_model_to_ref_name_map = {}
_ref_name_to_schema_map = {}


class BaseModel(object):
    def __init__(self, *args, **kwargs):
        processed_key_set = set()

        # load from scdhemas
        for name, field in self.iter_schema_fields():
            value = kwargs.get(name, None)

            position = [name]
            value = field.parse(value, position)

            setattr(self, name, value)
            processed_key_set.add(name)

        # load additionalProperties
        # TODO: add config class for dealing with additional properties
        # TODO: add root class setting
        for k, v in six.iteritems(kwargs):
            if k in processed_key_set:
                continue

            setattr(self, k, v)

    def __to_json_value(self, v):
        if isinstance(v, BaseModel):
            return v.to_json_dict()

        elif isinstance(v, Mapping):
            return {_k: self.__to_json_value(_v) for _k, _v in six.iteritems(v)}

        elif is_iterable(v):
            return [self.__to_json_value(x) for x in v]

        else:
            return v

    def to_json_dict(self):
        json_d = {}
        for k, v in self.__dict__.items():
            json_d[k] = self.__to_json_value(v)

        return json_d

    @classmethod
    def get_schema_field(cls, name):
        field = getattr(cls, name, None)
        if isinstance(field, BaseSchemaElement):
            return field

        return None

    @classmethod
    def iter_schema_fields(cls):
        for name in dir(cls):
            if name.startswith('__'):
                continue

            value = getattr(cls, name)
            if isinstance(value, BaseSchemaElement):
                yield (name, value)

    @classmethod
    def get_json_schema(cls):
        if cls in _model_to_ref_name_map:
            return _ref_name_to_schema_map[_model_to_ref_name_map[cls]]

        schema_d = {}
        schema_d['type'] = 'object'
        schema_d['properties'] = OrderedDict()
        schema_d['required'] = []

        for name, field in cls.iter_schema_fields():
            schema_d['properties'][name] = field.get_json_schema()
            if field.required:
                schema_d['required'].append(name)

        ref_name = cls.__name__
        if ref_name in _ref_name_to_schema_map:
            ref_name = cls.__module__.replace('.', '_') + '_' + ref_name

        _ref_name_to_schema_map[ref_name] = schema_d
        _model_to_ref_name_map[cls] = ref_name

        return schema_d

    @classmethod
    def get_json_schema_ref(cls):
        if cls in _model_to_ref_name_map:
            return _model_to_ref_name_map[cls]

        cls.get_json_schema()
        return _model_to_ref_name_map[cls]

    @classmethod
    def get_ref_name_to_schema_map(cls):
        return copy(_ref_name_to_schema_map)

    @classmethod
    def parse(cls, value):
        if isinstance(value, cls):
            return value

        elif isinstance(value, Mapping):
            value_dict = value

        else:
            value_dict = {
                k: getattr(value, k) for k in dir(value) if not k.startswith('_')
            }

        return cls(**value_dict)



# -*- coding:utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

import six
from collections import Iterable

from six import binary_type

_ALL_STRING_LIKE_TYPES = tuple(
    [six.text_type, six.binary_type] + list(six.string_types)
)


def int_or_none(value):
    return value if isinstance(value, int) else None


def ensure_set(values):
    if isinstance(values, _ALL_STRING_LIKE_TYPES):
        return set([values])

    if isinstance(values, Iterable):
        return set(values)


def is_iterable(value):
    return isinstance(value, Iterable) and not isinstance(
        value, _ALL_STRING_LIKE_TYPES
    )




# -*- coding:utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from abc import ABCMeta, abstractmethod

import six

from .constants import NO_DEFULAT_VALUE
from .exceptions import SchemaValidationError


class BaseSchemaElement(six.with_metaclass(ABCMeta, object)):
    @abstractmethod
    def __init__(
        self,
        default_value=NO_DEFULAT_VALUE,
        required=True,
        title=None,
        description=None,
        example=None,
    ):
        self.default_value = default_value
        self.required = required
        self.title = title
        self.description = description
        self.example = example

    @abstractmethod
    def parse(self, value, position):  # list of position info passed by caller
        if value is None:
            if self.default_value is not NO_DEFULAT_VALUE:
                value = self.default_value

        if value is None and self.required:
            raise SchemaValidationError(
                value, 'FIELD_IS_REQUIRED', position=position
            )

        return value

    @abstractmethod
    def get_json_schema(self):
        schema_d = {}
        if self.title:
            schema_d['title'] = self.title
        if self.description:
            schema_d['description'] = self.description
        if self.example:
            schema_d['examples'] = [self.example]
        if self.default_value is not NO_DEFULAT_VALUE:
            schema_d['default'] = self.default_value

        return schema_d

    # @abstractmethod
    # def get_value(self):
    #     pass

    # @abstractmethod
    # def get_json_value(self):
    #     pass



# -*- coding:utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from collections import Iterable

import six

from .base import BaseSchemaElement
from .constants import NO_DEFULAT_VALUE
from .exceptions import SchemaValidationError
from .utils import int_or_none


class ArrayField(BaseSchemaElement):
    def __init__(
        self,
        item_field,
        default_value=NO_DEFULAT_VALUE,
        required=True,
        title=None,
        description=None,
        example=None,
        min_items=None,
        max_items=None,
        check_unique_items=False,
    ):
        super(ArrayField, self).__init__(
            default_value=default_value,
            required=required,
            title=title,
            description=description,
            example=example,
        )

        assert isinstance(item_field, BaseSchemaElement)
        self.item_field = item_field

        self.min_items = int_or_none(min_items)
        self.max_items = int_or_none(max_items)
        self.check_unique_items = check_unique_items

    def parse(self, value, position):
        value = super(ArrayField, self).parse(value, position)

        if value is None and not self.required:
            return value

        if isinstance(value, six.text_type) or not isinstance(value, Iterable):
            raise SchemaValidationError(
                value, 'TYPE_MISMATCH', constraint='array', position=position,
            )

        if self.check_unique_items:
            value_set = set()
            for idx, item in enumerate(value):
                if item in value_set:
                    raise SchemaValidationError(
                        item,
                        'ARRAY_HAS_NON_UNIQUE_ITEM',
                        constraint=None,
                        position=position + [idx],
                    )
                value_set.add(item)

        new_values = []
        for idx, item in enumerate(value):
            new_values.append(self.item_field.parse(item, position + [idx]))

        value = new_values

        if self.min_items is not None and len(value) < self.min_items:
            raise SchemaValidationError(
                value,
                'ARRAY_HAS_TOO_FEW_ITEMS',
                constraint=self.min_items,
                position=position,
            )

        if self.max_items is not None and len(value) > self.max_items:
            raise SchemaValidationError(
                value,
                'ARRAY_HAS_TOO_MANY_ITEMS',
                constraint=self.max_items,
                position=position,
            )

        return value

    def get_json_schema(self):
        schema_d = super(ArrayField, self).get_json_schema()
        schema_d['type'] = 'array'
        schema_d['items'] = self.item_field.get_json_schema()
        if self.min_items is not None:
            schema_d['minItems'] = self.min_items
        if self.max_items is not None:
            schema_d['maxItems'] = self.max_items
        if self.check_unique_items:
            schema_d['uniqueItems'] = True
        return schema_d



# -*- coding:utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function


SCHEMA_ERROR_TYPE_SET = {
    'TYPE_MISMATCH',
    'VALUE_NOT_IN_ENUM',
    'VALUE_MUST_GREATER_THAN',
    'VALUE_MUST_GREATER_EQUAL_THAN',
    'VALUE_MUST_LESSER_THAN',
    'VALUE_MUST_LESSER_EQUAL_THAN',
    'VALUE_NOT_MUTLIPLE_OF',
    'ARRAY_HAS_TOO_FEW_ITEMS',
    'ARRAY_HAS_TOO_MANY_ITEMS',
    'ARRAY_HAS_NON_UNIQUE_ITEM',
    'TEXT_TOO_LONG',
    'TEXT_TOO_SHORT',
    'REGEX_NOT_MATCH',
    'FIELD_IS_REQUIRED',
}

NO_DEFULAT_VALUE = object()
NOT_SET = object()



# -*- coding:utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import

import sys
from collections import Mapping

from .base import BaseSchemaElement
from .constants import NO_DEFULAT_VALUE
from .exceptions import SchemaValidationError
from .utils import int_or_none


class ObjectField(BaseSchemaElement):
    def __init__(
        self,
        model_cls,
        default_value=NO_DEFULAT_VALUE,
        required=True,
        title=None,
        description=None,
        example=None,
        # additional_properties=True,
        # min_properties=None,
        # max_properties=None,
    ):
        from ..base import BaseModel

        assert isinstance(model_cls, type) and issubclass(model_cls, BaseModel)

        super(ObjectField, self).__init__(
            default_value=default_value,
            required=required,
            title=title,
            description=description,
            example=example,
        )
        self.model_cls = model_cls
        # self.additional_properties = additional_properties
        # self.min_properties = int_or_none(min_properties)
        # self.max_properties = int_or_none(max_properties)

    def parse(self, value, position):
        value = super(ObjectField, self).parse(value, position)

        if value is None and not self.required:
            return value

        try:
            value = self.model_cls.parse(value)
        except SchemaValidationError as e:
            e.position = position + e.position
            raise e

        return value

    def get_json_schema(self):
        schema_d = super(ObjectField, self).get_json_schema()
        schema_d.update(self.model_cls.get_json_schema())
        return schema_d



# -*- coding:utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

import six

from .base import BaseSchemaElement
from .constants import NO_DEFULAT_VALUE
from .exceptions import SchemaValidationError

from .utils import int_or_none, ensure_set

import re


class StringField(BaseSchemaElement):
    def __init__(
        self,
        default_value=NO_DEFULAT_VALUE,
        required=True,
        title=None,
        description=None,
        example=None,
        str_format=None,
        min_length=None,
        max_length=None,
        regex=None,
        enums=None,
    ):
        super(StringField, self).__init__(
            default_value=default_value,
            required=required,
            title=title,
            description=description,
            example=example,
        )
        self.format = str_format
        self.min_length = int_or_none(min_length)
        self.max_length = int_or_none(max_length)
        self.regex = re.compile(regex) if regex else None
        if enums:
            self.enums = tuple(enums)
            self.enum_set = set(self.enums)
        else:
            self.enums = None
            self.enum_set = None

    def parse(self, value, position):
        value = super(StringField, self).parse(value, position)

        if value is None and not self.required:
            return value

        try:
            value = six.ensure_text(value)
        except:
            raise SchemaValidationError(
                value, 'TYPE_MISMATCH', constraint='string', position=position
            )

        if self.min_length is not None and len(value) < self.min_length:
            raise SchemaValidationError(
                value,
                'TEXT_TOO_SHORT',
                constraint=self.min_length,
                position=position,
            )

        if self.max_length is not None and len(value) > self.max_length:
            raise SchemaValidationError(
                value,
                'TEXT_TOO_LONG',
                constraint=self.max_length,
                position=position,
            )

        if self.enums and value not in self.enum_set:
            raise SchemaValidationError(
                value,
                'VALUE_NOT_IN_ENUM',
                constraint=self.enums,
                position=position,
            )

        if self.regex and not self.regex.match(value):
            raise SchemaValidationError(
                value,
                'REGEX_NOT_MATCH',
                constraint=self.regex.pattern,
                position=position,
            )

        return value

    def get_json_schema(self):
        schema_d = super(StringField, self).get_json_schema()

        schema_d['type'] = 'string'
        if self.min_length is not None:
            schema_d['minLength'] = self.min_length
        if self.max_length is not None:
            schema_d['maxLength'] = self.max_length
        if self.enums:
            schema_d['enum'] = list(self.enums)
        if self.format:
            schema_d['format'] = self.format

        return schema_d



# -*- coding:utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function
import re

import six

from .base import BaseSchemaElement
from .constants import NO_DEFULAT_VALUE
from .exceptions import SchemaValidationError
from .utils import ensure_set


class BooleanField(BaseSchemaElement):
    def __init__(
        self,
        default_value=NO_DEFULAT_VALUE,
        required=True,
        title=None,
        description=None,
        example=None,
    ):
        super(BooleanField, self).__init__(
            default_value=default_value,
            required=required,
            title=title,
            description=description,
            example=example,
        )

    def parse(self, value, position):
        org_value = value = super(BooleanField, self).parse(value, position)

        if value is None and not self.required:
            return value

        if value in (True, False):
            return value

        if isinstance(value, six.integer_types):
            value = six.text_type(value)

        if isinstance(value, six.string_types):
            value = six.ensure_text(value).lower().strip()

            if value in ('true', 'y', 'yes', '1'):
                return True
            elif value in ('false', 'n', 'no', '0'):
                return False

        raise SchemaValidationError(
            org_value,
            'TYPE_MISMATCH',
            constraint='boolean',
            position=position,
        )

    def get_json_schema(self):
        schema_d = super(BooleanField, self).get_json_schema()

        schema_d['type'] = 'boolean'

        return schema_d



# -*- coding:utf8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from abc import ABCMeta, abstractmethod

import six
import json

from django.core.files.uploadedfile import UploadedFile

from .schema.fields import BaseSchemaElement, StringField
from .schema import BaseModel
from .schema.fields.exceptions import SchemaValidationError

_NOT_SET = object()


class BaseRequestParam(six.with_metaclass(ABCMeta, object)):
    IN_POS = _NOT_SET

    @abstractmethod
    def __init__(self):
        assert self.IN_POS is not _NOT_SET

    @abstractmethod
    def get_openapi_schema(self, name):
        pass

    @abstractmethod
    def get_value_from_request(self, request, name):
        pass

    @abstractmethod
    def parse(self, request, name):
        pass


class BaseRequestField(BaseRequestParam):
    def __init__(self, field=None):
        super(BaseRequestField, self).__init__()
        field = field or StringField()

        assert isinstance(field, BaseSchemaElement)
        self.field = field

    def get_openapi_schema(self, name):
        field_schema_d = self.field.get_json_schema()
        description_segs = [
            field_schema_d.pop(k)
            for k in ('title', 'description')
            if field_schema_d.get(k)
        ]

        schema_d = {
            'name': name,
            'required': self.field.required,
            'schema': self.field.get_json_schema(),
            'in': self.IN_POS,
        }

        if description_segs:
            schema_d['description'] = '\n'.join(description_segs)

        return schema_d

    def parse(self, request, name):
        return self.field.parse(
            self.get_value_from_request(request, name), [self.IN_POS, name]
        )


class BaseRequestBodyModel(BaseRequestParam):
    def __init__(self, model_cls):
        assert isinstance(model_cls, type) and issubclass(model_cls, BaseModel)
        self.model_cls = model_cls

    def get_openapi_schema(self):
        schema_d = {
            'content': {
                'application/json': {
                    'schema': {
                        '$ref': '#/components/schemas/'
                        + self.model_cls.get_json_schema_ref()
                    }
                }
            },
            'required': True,
        }

        return schema_d

    def parse(self, request, name):
        try:
            return self.model_cls(**self.get_value_from_request(request, name))
        except SchemaValidationError as e:
            e.position = [self.IN_POS, name] + e.position
            raise e


class Body(BaseRequestBodyModel):
    IN_POS = 'body'

    def get_value_from_request(self, request, name):
        return json.loads(six.ensure_text(request.body))


class Query(BaseRequestField):
    IN_POS = 'query'

    def get_value_from_request(self, request, name):
        return request.GET.get(name)


class Path(BaseRequestField):
    IN_POS = 'path'

    def get_value_from_request(self, request, name):
        return request.path_kwargs.get(name)


class Header(BaseRequestField):
    IN_POS = 'header'

    def get_value_from_request(self, request, name):
        return request.META.get(name.upper())


class Cookie(BaseRequestField):
    IN_POS = 'cookie'

    def get_value_from_request(self, request, name):
        return request.COOKIES.get(name)


class Form(BaseRequestField):
    IN_POS = 'form'

    def get_value_from_request(self, request, name):
        return request.POST.get(name)


class UploadFile(BaseRequestField):
    IN_POS = 'form'

    def __init__(self):
        super(UploadFile, self).__init__(
            StringField(title='Upload file', str_format='binary')
        )

    def get_value_from_request(self, request, name):
        return request.FILES.get(name)

    def parse(self, request, name):
        f = self.get_value_from_request(request, name)

        if not isinstance(f, UploadedFile):
            f = None

        if f is None:
            self.field.parse(f, [self.IN_POS, name])

        return f



from django import VERSION as DJANGO_VERSION

_SUPPORT_SAMESITE_COOKIE = (DJANGO_VERSION[0], DJANGO_VERSION[1]) >= (2, 1)


class CookieJar(object):
    def __init__(self):
        self.cookie_operations = []

    def _check_samesite_support(self, samesite):
        if samesite is not None and not _SUPPORT_SAMESITE_COOKIE:
            raise ValueError('samesite cookie is not support in your django version')

    def set_cookie(
        self,
        key,
        value='',
        max_age=None,
        expires=None,
        path='/',
        domain=None,
        secure=False,
        httponly=False,
        samesite=None,
    ):
        self._check_samesite_support(samesite)
        arg_d = dict(
            key=key,
            value=value,
            max_age=max_age,
            expires=expires,
            path=path,
            domain=domain,
            secure=secure,
            httponly=httponly,
        )
        if _SUPPORT_SAMESITE_COOKIE:
            arg_d['samesite'] = samesite

        self.cookie_operations.append(('set_cookie', arg_d))

    def delete_cookie(self, key, path='/', domain=None, samesite=None):
        self._check_samesite_support(samesite)
        arg_d = dict(key=key, path=path, domain=domain,)
        if _SUPPORT_SAMESITE_COOKIE:
            arg_d['samesite'] = samesite

        self.cookie_operations.append(('delete_cookie', arg_d))

    def apply_to_response(self, response):
        op_type_to_fn_map = {}

        for op_type, arg_d in self.cookie_operations:
            if op_type not in op_type_to_fn_map:
                op_type_to_fn_map[op_type] = getattr(response, op_type)

            fn = op_type_to_fn_map[op_type]
            fn(**arg_d)



# -*- coding:utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from .constants import SCHEMA_ERROR_TYPE_SET


class SchemaValidationError(Exception):
    def __init__(self, value, err_type, constraint=None, position=None):
        self.value = value
        assert err_type in SCHEMA_ERROR_TYPE_SET
        self.err_type = err_type
        self.constraint = constraint
        self.position = position or []

    def __repr__(self):
        return '{}(value={}, err_type={}, constraint={}, position={})'.format(
            self.__class__.__name__,
            repr(self.value),
            repr(self.err_type),
            repr(self.constraint),
            repr(self.position),
        )

    def __str__(self):
        return self.__repr__()



# -*- coding:utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from django.http import JsonResponse


def json_response(data, status_code=200):
    resp = JsonResponse(data)
    resp.status_code = status_code
    return resp



# -*- coding:utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from .base import BaseSchemaElement
from .constants import NO_DEFULAT_VALUE
from .exceptions import SchemaValidationError
from .utils import ensure_set


class NumberField(BaseSchemaElement):
    def __init__(
        self,
        default_value=NO_DEFULAT_VALUE,
        required=True,
        title=None,
        description=None,
        example=None,
        gt=None,
        gte=None,
        lt=None,
        lte=None,
        multiple_of=None,
        enums=None,
    ):
        super(NumberField, self).__init__(
            default_value=default_value,
            required=required,
            title=title,
            description=description,
            example=example,
        )
        self.gt = gt
        self.gte = gte
        self.lt = lt
        self.lte = lte
        self.multiple_of = multiple_of
        if enums:
            self.enums = tuple(enums)
            self.enum_set = set(self.enums)
        else:
            self.enums = None
            self.enum_set = None

    def parse(self, value, position):
        value = super(NumberField, self).parse(value, position)

        if value is None and not self.required:
            return value

        if not isinstance(value, (int, float)):
            try:
                f_value = float(value)
                value = int(value) if f_value.is_integer() else f_value
            except (TypeError, ValueError):
                raise SchemaValidationError(
                    value, 'TYPE_MISMATCH', constraint='number', position=position,
                )

        if self.enums and value not in self.enum_set:
            raise SchemaValidationError(
                value, 'VALUE_NOT_IN_ENUM', constraint=self.enums, position=position,
            )

        if self.gt is not None and not value > self.gt:
            raise SchemaValidationError(
                value, 'VALUE_MUST_GREATER_THAN', constraint=self.gt, position=position,
            )

        if self.gte is not None and not value >= self.gte:
            raise SchemaValidationError(
                value,
                'VALUE_MUST_GREATER_EQUAL_THAN',
                constraint=self.gte,
                position=position,
            )

        if self.lt is not None and not value < self.lt:
            raise SchemaValidationError(
                value, 'VALUE_MUST_LESSER_THAN', constraint=self.lt, position=position,
            )

        if self.lte is not None and not value <= self.lte:
            raise SchemaValidationError(
                value,
                'VALUE_MUST_LESSER_EQUAL_THAN',
                constraint=self.gte,
                position=position,
            )

        if self.multiple_of is not None and value % self.multiple_of != 0:
            raise SchemaValidationError(
                value,
                'VALUE_NOT_MUTLIPLE_OF',
                constraint=self.multiple_of,
                position=position,
            )

        return value

    def get_json_schema(self):
        schema_d = super(NumberField, self).get_json_schema()

        schema_d['type'] = 'number'
        if self.gt is not None:
            schema_d['exclusiveMinimum'] = self.gt
        if self.gte is not None:
            schema_d['minimum'] = self.gte
        if self.lt is not None:
            schema_d['exclusiveMaximum'] = self.lt
        if self.lte is not None:
            schema_d['maximum'] = self.lte

        return schema_d



# -*- coding:utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from inspect import getargspec
from string import Formatter
from collections import OrderedDict
import re

from django.http import HttpResponse

from .utils import json_response
from .params import BaseRequestParam
from .cookie import CookieJar
from .schema import (
    BaseModel,
    StringField,
    ArrayField,
    ObjectField,
    SchemaValidationError,
)

from collections import Counter, defaultdict, OrderedDict

import six


ALLOW_HTTP_METHOD_SET = {
    'GET',
    'POST',
    'PUT',
    'DELETE',
    'PATCH',
    'HEAD',
    'TRACE',
}


class RouteConfig(BaseModel):
    route_path = StringField(required=True, min_length=1, regex=r'^/')
    allow_methods = ArrayField(StringField(enums=ALLOW_HTTP_METHOD_SET), min_items=1)
    summary = StringField(default_value=None, required=False)
    description = StringField(default_value=None, required=False)
    tags = ArrayField(StringField())


class ValidationErrorItem(BaseModel):
    loc = ArrayField(StringField())
    msg = StringField()
    type = StringField()


class ValidationErrorResponse(BaseModel):
    deatil = ArrayField(ObjectField(ValidationErrorItem))


class PathNotFullfilled(Exception):
    pass


PATH_NOT_FULL_FILLED = object()


class RoutePath(object):
    def __init__(self, route_path):
        route_path = six.ensure_text(route_path)
        assert route_path.startswith('/')

        self.org_route_path = route_path
        # route_path = route_path[1:]
        self.route_path = route_path[:-1] if route_path.endswith('/') else route_path

        self.key_set = set()
        re_segs = []
        fmt = Formatter()
        for prefix, key, fmt_spec, conversion in fmt.parse(self.route_path):
            if key is None:
                re_segs.append(prefix)
                continue

            assert key and not fmt_spec and not conversion, (
                'fail parsing parameter from path: ' + self.route_path
            )
            assert key not in self.key_set, 'duplicated key {}'.format(key)
            self.key_set.add(key)
            re_segs.append('{prefix}(?P<{key}>[^/]+)'.format(prefix=prefix, key=key))

        # print('regex: ' + ''.join(re_segs))
        self.regex = re.compile('^' + ''.join(re_segs) + '$')

    def parse(self, request_path):
        if not request_path.startswith('/'):
            request_path = '/' + request_path

        match = self.regex.match(request_path)
        if not match:
            return PATH_NOT_FULL_FILLED

        return match.groupdict()


class Route(object):
    def __init__(
        self,
        fn,
        route_path,
        allow_methods=None,
        summary=None,
        description=None,
        tags=tuple(),
        response_model=None,
        response_model_map=None,
    ):
        description = description or fn.__doc__

        cfg = RouteConfig(
            route_path=route_path,
            allow_methods=allow_methods,
            summary=summary,
            description=description,
            tags=tags,
        )

        if route_path.endswith('/'):
            route_path = route_path[:-1]

        self.route_path = route_path
        self.path_parser = RoutePath(route_path)
        self.allow_methods = cfg.allow_methods
        self.summary = cfg.summary
        self.description = cfg.description
        self.tags = cfg.tags

        self.response_map = {200: BaseModel, 422: ValidationErrorResponse}
        if response_model:
            assert isinstance(response_model, type) and issubclass(
                response_model, BaseModel
            )
            self.response_map[200] = response_model

        if response_model_map and isinstance(response_model_map, dict):
            for k, v in response_model_map.items():
                assert (
                    isinstance(k, int) and 100 <= k < 1000
                ), 'Bad HTTP status code {}'.format(k)
                assert isinstance(v, type) and issubclass(
                    v, BaseModel
                ), 'Must be OpenAPI BaseModel'
                response_model_map[k] = v

        self.arg_type_counter = Counter()
        self.arg_name_to_request_param_map = OrderedDict()
        self.pass_request = False
        self.pass_session = False
        self.pass_cookie_jar = False
        self._body_form_cls = None

        arg_spec = getargspec(fn)
        if arg_spec.args:
            arg_default_len_diff = len(arg_spec.args) - (
                len(arg_spec.defaults) if arg_spec.defaults else 0
            )
            for i in range(len(arg_spec.args)):
                name = arg_spec.args[i]
                vidx = i - arg_default_len_diff
                value = arg_spec.defaults[vidx] if vidx >= 0 else None

                # print(name, value)

                if name == 'request':
                    self.pass_request = True

                elif name == 'session':
                    self.pass_session = True

                elif name == 'cookie_jar':
                    self.pass_cookie_jar = True

                elif isinstance(value, BaseRequestParam):
                    assert (
                        name not in self.arg_name_to_request_param_map
                    ), 'duplicated arg name'
                    self.arg_name_to_request_param_map[name] = value
                    self.arg_type_counter[value.IN_POS] += 1

                else:
                    raise ValueError('unmapped parameter {}'.format(name))

        if self.path_parser.key_set:
            for k in self.path_parser.key_set:
                assert (
                    k in self.arg_name_to_request_param_map
                    and self.arg_name_to_request_param_map[k].IN_POS == 'path'
                ), 'umapped path arg key {}'.format(k)

        assert (
            self.arg_type_counter['body'] <= 1
        ), 'can only has single Body param in same route'

        assert (
            self.arg_type_counter['body'] == 0 or self.arg_type_counter['form'] == 0
        ), 'can not define both form/body param in same route'

        self.fn = fn

    def get_openapi_schema(self):
        route_d = {}

        if self.summary:
            route_d['summary'] = self.summary
        if self.description:
            route_d['description'] = self.description
        if self.tags:
            route_d['tags'] = self.tags

        form_param_d = OrderedDict()
        parameters = []
        for name, param in six.iteritems(self.arg_name_to_request_param_map):
            if param.IN_POS == 'body':
                route_d['requestBody'] = param.get_openapi_schema()

            elif param.IN_POS == 'form':
                form_param_d[name] = param.field

            else:
                parameters.append(param.get_openapi_schema(name))

        if parameters:
            route_d['parameters'] = parameters

        if form_param_d and 'requestBody' not in route_d:
            if self._body_form_cls is None:
                self._body_form_cls = type(
                    six.ensure_str(
                        '{route_path}_body_form'.format(
                            route_path=self.route_path.replace('/', '_')
                        )
                    ),
                    (BaseModel,),
                    form_param_d,
                )

            route_d['requestBody'] = {
                'content': {
                    'multipart/form-data': {
                        # 'application/x-www-form-urlencoded': {
                        'schema': {
                            '$ref': '#/components/schemas/'
                            + self._body_form_cls.get_json_schema_ref()
                        }
                    }
                }
            }

        if self.response_map:
            route_d['responses'] = {
                six.text_type(k): {
                    'content': {
                        'application/json': {
                            'schema': {
                                '$ref': '#/components/schemas/'
                                + v.get_json_schema_ref()
                            }
                        }
                    }
                }
                for k, v in self.response_map.items()
            }

        json_d = {m.lower(): route_d for m in self.allow_methods}
        return json_d

    def match_path(self, request_path):
        return self.path_parser.parse(request_path)

    def prase_response(self, resp, http_status_code=200):

        resp = resp or {}  # default empty dict response
        resp_model_cls = self.response_map.get(http_status_code)

        if isinstance(resp, HttpResponse):
            return resp

        if isinstance(resp, dict):
            if resp_model_cls:
                resp = resp_model_cls(**resp)
            else:
                return json_response(resp)

        if isinstance(resp, BaseModel) and (
            not resp_model_cls or isinstance(resp, resp_model_cls)
        ):
            return json_response(resp.to_json_dict())

        raise ValueError(
            'unable to process resp of {route_path}'.format(route_path=self.route_path)
        )

    def __call__(self, request):
        kwargs = {}

        if self.pass_request:
            kwargs['request'] = request

        if self.pass_session:
            kwargs['session'] = getattr(request, 'session', None)

        if self.pass_cookie_jar:
            kwargs['cookie_jar'] = CookieJar()

        validation_errors = []
        for k, field in six.iteritems(self.arg_name_to_request_param_map):
            try:
                kwargs[k] = field.parse(request, k)
            except SchemaValidationError as e:
                validation_errors.append(e)

        if validation_errors:
            return json_response(
                {
                    'detail': [
                        {
                            'loc': err.position,
                            'type': err.err_type,
                            'msg': err.err_type,
                        }
                        for err in validation_errors
                    ]
                },
                status_code=422,
            )

        resp = self.fn(**kwargs)
        resp = self.prase_response(resp)

        if self.pass_cookie_jar:
            kwargs['cookie_jar'].apply_to_response(resp)

        return resp

    def __repr__(self):
        return 'Route("{}")'.format(self.route_path)



# -*- coding:utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import setuptools

with open('README.md') as f:
    long_description = f.read()

setuptools.setup(
    name='django-openapi',
    version='0.1.1.9',
    author='toki kanno',
    author_email='toki.kanno@gmail.com',
    description='A minimal FastAPI implementation in python2 + Django without pydantic',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/tokikanno/django-openapi',
    packages=setuptools.find_packages(
        '.', include=('django_openapi', 'django_openapi.*',)
    ),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    python_requires='>=2.7',
    setup_requires=['wheel'],
    install_requires=['django', 'six'],
)



# -*- coding:utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from collections import OrderedDict

import six

from django.conf.urls import url
from django.views.decorators.csrf import csrf_exempt
from django.http import HttpResponse, Http404, HttpResponseNotAllowed

from .route import Route, PATH_NOT_FULL_FILLED
from .schema import BaseModel

from .utils import json_response

DOC_PAGE_TPL = '''<!DOCTYPE html>
<html>
<head>
    <link type='text/css' rel='stylesheet' href='https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui.css'>
    <title>{title} - Swagger UI</title>
</head>
<body>
    <div id='swagger-ui'>
    </div>
    <script src='https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui-bundle.js'></script>
    <!-- `SwaggerUIBundle` is now available on the page -->
    <script>
    const ui = SwaggerUIBundle({{
        url: '/{prefix_path}/_openapi.json',
        oauth2RedirectUrl: window.location.origin + '/docs/oauth2-redirect',
        dom_id: '#swagger-ui',
        presets: [
            SwaggerUIBundle.presets.apis,
            SwaggerUIBundle.SwaggerUIStandalonePreset
        ],
        layout: 'BaseLayout',
        deepLinking: true
    }})
    </script>
</body>
</html>'''

REDOC_PAGE_TPL = '''
<!DOCTYPE html>
<html>
    <head>
        <title>{title} - ReDoc</title>
        <!-- needed for adaptive design -->
        <meta charset='utf-8'/>
        <meta name='viewport' content='width=device-width, initial-scale=1'>

        <link href='https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700' rel='stylesheet'>

        <!--
        ReDoc doesn't change outer page styles
        -->
        <style>
        body {{
            margin: 0;
            padding: 0;
        }}
        </style>
    </head>
    <body>
    <redoc spec-url='/{prefix_path}/_openapi.json'></redoc>
    <script src='https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js'> </script>
    </body>
</html>'''

RAPIDOC_PAGE_TPL = '''
<!doctype html> <!-- Important: must specify -->
<html>
<head>
  <title>{title} - RapiDoc</title>
  <meta charset="utf-8"> <!-- Important: rapi-doc uses utf8 charecters -->
  <script type="module" src="https://unpkg.com/rapidoc/dist/rapidoc-min.js"></script>
</head>
<body>
  <rapi-doc
    spec-url="/{prefix_path}/_openapi.json"
    sort-endpoints-by="method"
    render-style="read"
  > </rapi-doc>
</body>
</html>
'''


class OpenAPI(object):
    def __init__(
        self,
        title='OpenAPI',  # REQUIRED. The title of the API.
        version='0.0',  # REQUIRED. The version of the OpenAPI document
        description='',  # A short description of the API. CommonMark syntax
        prefix_path='',
        server_url='',
        server_description='',
    ):
        self.title = title
        self.description = description
        self.version = version
        self.route_path_set = set()
        self.routes = []
        self.prefix_path = prefix_path.strip('/')
        self.server_info_d = {
            'url': server_url,
            'description': server_description,
        }

    def add_route(
        self,
        route_path,
        allow_methods=tuple(),
        summary='',
        description='',
        tags=tuple(),
        response_model=None,
        response_model_map=None,
    ):
        def _decorator(fn):
            route = Route(
                fn,
                route_path=route_path,
                allow_methods=allow_methods,
                summary=summary,
                description=description,
                tags=tags,
                response_model=response_model,
                response_model_map=response_model_map,
            )

            assert (
                route_path not in self.route_path_set
            ), '{} already registered'.format(route_path)

            self.routes.append(route)

            return fn

        return _decorator

    def get(
        self,
        route_path,
        summary='',
        description='',
        tags=tuple(),
        response_model=None,
        response_model_map=None,
    ):
        return self.add_route(
            route_path=route_path,
            allow_methods=['GET'],
            summary=summary,
            description=description,
            tags=tags,
            response_model=response_model,
            response_model_map=response_model_map,
        )

    def post(
        self,
        route_path,
        summary='',
        description='',
        tags=tuple(),
        response_model=None,
        response_model_map=None,
    ):
        return self.add_route(
            route_path=route_path,
            allow_methods=['POST'],
            summary=summary,
            description=description,
            tags=tags,
            response_model=response_model,
            response_model_map=response_model_map,
        )

    def get_openapi_schema(self):
        api_d = {
            'openapi': '3.0.2',
            'info': {
                'title': self.title,
                'description': self.description,
                'version': self.version,
            },
        }

        if self.server_info_d['url']:
            api_d['servers'] = [self.server_info_d]

        api_d['paths'] = path_d = OrderedDict()

        for route_obj in self.routes:
            route_path = '/{prefix_path}{route_path}'.format(
                prefix_path=self.prefix_path, route_path=route_obj.route_path
            )

            if route_path not in api_d['paths']:
                api_d['paths'][route_path] = OrderedDict()

            api_d['paths'][route_path].update(route_obj.get_openapi_schema())

        api_d['components'] = {'schemas': BaseModel.get_ref_name_to_schema_map()}

        return api_d

    def as_django_url_pattern(self):
        return url(
            '^{prefix_path}(?P<route_path>/.*)'.format(prefix_path=self.prefix_path),
            self.as_django_view(),
        )

    def as_django_view(self):
        @csrf_exempt
        def dispatcher(request, route_path):
            route_path = route_path[:-1] if route_path.endswith('/') else route_path

            # document routes
            if route_path == '/_openapi.json':
                return json_response(self.get_openapi_schema())
            elif route_path == '/_docs':
                return HttpResponse(
                    DOC_PAGE_TPL.format(title=self.title, prefix_path=self.prefix_path)
                )
            elif route_path == '/_redoc':
                return HttpResponse(
                    REDOC_PAGE_TPL.format(
                        title=self.title, prefix_path=self.prefix_path
                    )
                )
            elif route_path == '/_rapidoc':
                return HttpResponse(
                    RAPIDOC_PAGE_TPL.format(
                        title=self.title, prefix_path=self.prefix_path
                    )
                )

            # find route handler from path_route_map
            matched_route_to_path_kwargs_map = OrderedDict()

            for route in self.routes:
                path_kwargs = route.match_path(route_path)
                if path_kwargs is PATH_NOT_FULL_FILLED:
                    continue
                matched_route_to_path_kwargs_map[route] = path_kwargs

            # not route matched
            if not matched_route_to_path_kwargs_map:
                raise Http404

            for route, path_kwargs in six.iteritems(matched_route_to_path_kwargs_map):
                if request.method in route.allow_methods:
                    request.path_kwargs = path_kwargs
                    return route(request)
            else:
                raise HttpResponseNotAllowed

        return dispatcher
